#include "vmm_tbl.h"

/** Constructor for the VMMTbl class */
VMMTbl::VMMTbl(RMM& rmm, int ring, int node, int hybrids)
  : FrontEndBase(rmm.get_rmm_reg_access(), ring, node),
    pVMMConfig(*this),
    pRMM(&rmm)
{
  number_hybrids = hybrids;
  addr_list = ring * NUMBER_NODES_PER_RING + node;
  fe_addr_map[ring] = vmm_tbl_register_map;
  createEpicsParams();
  setInitialEpicsParams();
  pRMM->get_rmm_rings().set_topology(ring, node, "vmm-tbl");
  pRMM->updateTopologyPvs(ring, node, "vmmTbl");
  for (int hyb = 0; hyb < number_hybrids; hyb++) {
    pVMMConfig.enableHybrid(hyb, true);
    pRMM->setStringParam(addr_list, vmmHybFwVersion_[hyb], pVMMConfig.readFwVersion(hyb).c_str());
    pRMM->setStringParam(addr_list, vmmHybId_[hyb], pVMMConfig.readIDChip(hyb).c_str());
    pRMM->setStringParam(addr_list, vmmHybGeoPos_[hyb], pVMMConfig.readGeoPos(hyb).c_str());
    pRMM->setIntegerParam(addr_list, vmmHybLinkStatus_[hyb], std::stoi(pVMMConfig.checkLinkStatus(hyb)));
  }
}

void VMMTbl::configFE() {
  puts("config_FE");
}

void VMMTbl::setInitialEpicsParams() {
  puts("set_initial_epics_params");
}

asynStatus VMMTbl::createEpicsParams() {
  pRMM->createParam(addr_list, "SEL_ANALOG_MONITOR_VMM0", asynParamInt32, &vmmSelectMonitorVMM0_);
  pRMM->createParam(addr_list, "SEL_ANALOG_MONITOR_VMM1", asynParamInt32, &vmmSelectMonitorVMM1_);
  pRMM->createParam(addr_list, "READ_ADC_VMMS", asynParamInt32, &vmmProcessReadMonitorVMMs_);
  pRMM->createParam(addr_list, "ADC_VALUE_VMM0", asynParamFloat64, &vmmMonitorValueVMM0_);
  pRMM->createParam(addr_list, "ADC_VALUE_VMM1", asynParamFloat64, &vmmMonitorValueVMM1_);
  pRMM->createParam(addr_list, "VMM_FEN_ACQUIRE", asynParamInt32, &vmmAcquire_);

  for (int hyb = 0; hyb < number_hybrids; hyb++) {
    std::ostringstream param_name;
    param_name << "HYB_" << hyb <<"_FW_VERSION";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamOctet, &vmmHybFwVersion_[hyb]);
    param_name.str("");
    param_name << "HYB_" << hyb <<"_ID";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamOctet, &vmmHybId_[hyb]);
    param_name.str("");
    param_name << "HYB_" << hyb <<"_GEOPOS";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamOctet, &vmmHybGeoPos_[hyb]);
    param_name.str("");
    param_name << "HYB_" << hyb <<"_LINK_STATUS";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamInt32, &vmmHybLinkStatus_[hyb]);
    for (int vmm=0; vmm < VMMS_PER_HYBRID; vmm++)
    {
    param_name.str("");
    param_name << "HYB_" << hyb <<"_" << vmm << "_ST";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamInt8Array, &vmmST_[hyb][vmm]);
    param_name.str("");
    param_name << "HYB_" << hyb <<"_" << vmm << "_SC";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamInt8Array, &vmmSC_[hyb][vmm]);
    param_name.str("");
    param_name << "HYB_" << hyb <<"_" << vmm << "_SL";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamInt8Array, &vmmSL_[hyb][vmm]);
    param_name.str("");
    param_name << "HYB_" << hyb <<"_" << vmm << "_STH";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamInt8Array, &vmmSTH_[hyb][vmm]);
    param_name.str("");
    param_name << "HYB_" << hyb <<"_" << vmm << "_SM";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamInt8Array, &vmmSM_[hyb][vmm]);
    param_name.str("");
    param_name << "HYB_" << hyb <<"_" << vmm << "_SMX";
    pRMM->createParam(addr_list, param_name.str().c_str(), asynParamInt8Array, &vmmSMX_[hyb][vmm]);
    }
  }

  return asynSuccess;
}

asynStatus VMMTbl::writeInt32(asynUser *pasynUser, epicsInt32 value) {
  asynStatus status = asynSuccess;
  int function = pasynUser->reason;
  int addr = 0;
  pRMM->getAddress(pasynUser, &addr);

  if ((value) && (function == vmmProcessReadMonitorVMMs_)) {
    int adc_sensor = 0;
    for (int vmm = 0; vmm <= 1; vmm++) {
      pRMM->getIntegerParam(addr, vmmSelectMonitorVMM0_ + vmm, &adc_sensor);
      std::cout << "adc_sensor" << adc_sensor << std::endl;
      pVMMConfig.getHybrid(0).getVMM(vmm).setRegister("sm5_sm0", adc_sensor + 64);
      pVMMConfig.configVMM(0, vmm);
      int adc_read = pVMMConfig.readADC(0, vmm);

      std::cout << "adc_read" << adc_read << std::endl;

      if (adc_sensor == 3) {
        double temperature = (725 - adc_read) / 1.85;
        pRMM->setDoubleParam(addr, vmmMonitorValueVMM0_ + vmm, temperature);
      }
      else {
        pRMM->setDoubleParam(addr, vmmMonitorValueVMM0_ + vmm, adc_read);
      }
    }
  }

  else if (function == vmmAcquire_) {
  
    if (value) pVMMConfig.acquire(true);
    else pVMMConfig.acquire(false);

    for (int hyb = 0; hyb < number_hybrids; hyb++)
      pRMM->setIntegerParam(addr_list, vmmHybLinkStatus_[hyb], std::stoi(pVMMConfig.checkLinkStatus(hyb)));
  }

  return status;
}

asynStatus VMMTbl::writeInt8Array(asynUser *pasynUser, epicsInt8 *value, size_t nElements)
{

  int function, addr;
  int ncopy = CHANNELS_PER_VMM;
  const char *paramName;
  const char *functionName = "writeInt8Array";
  int status = asynSuccess;
  std::vector<epicsInt8> int8Array;
  Reg pair;

  status |= pRMM->parseAsynUser(pasynUser, &function, &addr, &paramName);
  pRMM->getParamName(addr, function, &paramName);
  status |= findRegister(function, &pair);
  if (status) {
    asynPrint(pasynUser, ASYN_TRACE_ERROR,
              "%s:%s: function=%d, name=%s, value=%d. Cannot find which Hybrid-VMM pair is function related to.\n",
              driverName, functionName, function, paramName, value);
    return (asynStatus) status;
  }

  asynPrint(pasynUser, ASYN_TRACEIO_DRIVER,
            "%s:%s: function=%d, name=%s, value=%d, hybrid=%d, vmm=%d, register=%s\n",
            driverName, functionName, function, paramName, value, pair.hyb, pair.vmm, pair.rg.c_str());

  for (epicsInt8 i = 0; i < ncopy; ++i) {
    int8Array.push_back(static_cast<int>(value[i]) >= 1 ? 1 : 0);
    pVMMConfig.getHybrid(pair.hyb).getVMM(pair.vmm).setRegister(pair.rg.c_str(),
                                                  static_cast<int>(value[i]),
                                                  i);
  }

  status |= pRMM->doCallbacksInt8Array(int8Array.data(), int8Array.size(), function, 0);
  return (asynStatus) status;

}

asynStatus VMMTbl::findRegister(int function, Reg *reg_){

  int (*array)[VMMS_PER_HYBRID];
  std::map<std::string, int(*)[VMMS_PER_HYBRID]>::iterator it;

  it = this->registers.begin();

  while ( it != this->registers.end() ){
    std::string reg_name = it->first;
    array = registers[reg_name];

    for (int h = 0; h < number_hybrids; h++){
      for (int v = 0; v < VMMS_PER_HYBRID; v++)
      {
        if (array[h][v] == function){
          reg_->hyb = h;
          reg_->vmm = v;
          reg_->rg = reg_name;
          return asynSuccess;
        }
      }
    }

    it++;
  }

  return asynError;

}

asynStatus VMMTbl::writeFloat64(asynUser *pasynUser, epicsFloat64 value) {
  asynStatus status = asynSuccess;
  int function = pasynUser->reason;
  int addr = 0;
  pRMM->getAddress(pasynUser, &addr);

  return status;
}


extern "C" {
  static void VMMTblConfig(const char *rmm_port_name, int ring, int hybrids) {
    int node = 0; // VMM use always only the Node 0
    RMM *pRMM;
    pRMM = (RMM *) findAsynPortDriver(rmm_port_name);
    pRMM->frontends[ring][node] = std::make_shared<VMMTbl>(*pRMM, ring, node, hybrids);
  }

  static const iocshArg configArg0 = { "RMM Port name", iocshArgString};
  static const iocshArg configArg1 = { "Ring", iocshArgInt};
  static const iocshArg configArg2 = { "Hybrids", iocshArgInt};


  static const iocshArg * const configArgs[] = {&configArg0,
                                                &configArg1,
                                                &configArg2};

  static const iocshFuncDef configFuncDef = {"VMMTblConfig", 3, configArgs};

  static void configCallFunc(const iocshArgBuf *args) {
    VMMTblConfig(args[0].sval, args[1].ival, args[2].ival);
  }

  static void VMMTblRegister(void) {
    iocshRegister(&configFuncDef, configCallFunc);
  }

  epicsExportRegistrar(VMMTblRegister);
} // end extern "C"